package io.quarkus.reactivemessaging.http.runtime;

import java.util.Optional;
import org.eclipse.microprofile.config.Config;

/**
 * Extract the outgoing configuration for the {@code quarkus-http} connector.
*/
public class QuarkusHttpConnectorOutgoingConfiguration extends QuarkusHttpConnectorCommonConfiguration {

  /**
   * Creates a new QuarkusHttpConnectorOutgoingConfiguration.
   */
  public QuarkusHttpConnectorOutgoingConfiguration(Config config) {
    super(config);
    validate();
  }

  /**
  * Gets the url value from the configuration.
  * Attribute Name: url
  * Description: The target URL
  * Mandatory: yes
  * @return the url
  */
  public String getUrl() {
    return config.getOptionalValue("url", String.class)
        .orElseThrow(() -> new IllegalArgumentException("The attribute `url` on connector 'quarkus-http' (channel: " + getChannel() + ") must be set"));
  }

  /**
  * Gets the serializer value from the configuration.
  * Attribute Name: serializer
  * Description: Message serializer
  * @return the serializer
  */
  public Optional<String> getSerializer() {
    return config.getOptionalValue("serializer", String.class);
  }

  /**
  * Gets the maxPoolSize value from the configuration.
  * Attribute Name: maxPoolSize
  * Description: Maximum pool size for connections
  * @return the maxPoolSize
  */
  public Optional<Integer> getMaxPoolSize() {
    return config.getOptionalValue("maxPoolSize", Integer.class);
  }

  /**
  * Gets the maxWaitQueueSize value from the configuration.
  * Attribute Name: maxWaitQueueSize
  * Description: Maximum requests allowed in the wait queue of the underlying client.  If the value is set to a negative number then the queue will be unbounded
  * @return the maxWaitQueueSize
  */
  public Optional<Integer> getMaxWaitQueueSize() {
    return config.getOptionalValue("maxWaitQueueSize", Integer.class);
  }

  /**
  * Gets the maxRetries value from the configuration.
  * Attribute Name: maxRetries
  * Description: The number of attempts to make for sending a request to a remote endpoint. Must not be less than zero
  * Default Value: 0
  * @return the maxRetries
  */
  public Integer getMaxRetries() {
    return config.getOptionalValue("maxRetries", Integer.class)
     .orElse(Integer.valueOf("0"));
  }

  /**
  * Gets the jitter value from the configuration.
  * Attribute Name: jitter
  * Description: Configures the random factor when using back-off with maxRetries > 0
  * Default Value: 0.5
  * @return the jitter
  */
  public String getJitter() {
    return config.getOptionalValue("jitter", String.class)
     .orElse("0.5");
  }

  /**
  * Gets the delay value from the configuration.
  * Attribute Name: delay
  * Description: Configures a back-off delay between attempts to send a request. A random factor (jitter) is applied to increase the delay when several failures happen.
  * @return the delay
  */
  public Optional<String> getDelay() {
    return config.getOptionalValue("delay", String.class);
  }

  public void validate() {
    super.validate();
    getUrl();
  }
}
