package io.quarkus.reactivemessaging.http.runtime;

import java.util.Optional;
import org.eclipse.microprofile.config.Config;

/**
 * Extract the outgoing configuration for the {@code quarkus-websocket} connector.
*/
public class QuarkusWebSocketConnectorOutgoingConfiguration extends QuarkusWebSocketConnectorCommonConfiguration {

  /**
   * Creates a new QuarkusWebSocketConnectorOutgoingConfiguration.
   */
  public QuarkusWebSocketConnectorOutgoingConfiguration(Config config) {
    super(config);
    validate();
  }

  /**
  * Gets the url value from the configuration.
  * Attribute Name: url
  * Description: The target URL
  * Mandatory: yes
  * @return the url
  */
  public String getUrl() {
    return config.getOptionalValue("url", String.class)
        .orElseThrow(() -> new IllegalArgumentException("The attribute `url` on connector 'quarkus-websocket' (channel: " + getChannel() + ") must be set"));
  }

  /**
  * Gets the serializer value from the configuration.
  * Attribute Name: serializer
  * Description: Message serializer
  * @return the serializer
  */
  public Optional<String> getSerializer() {
    return config.getOptionalValue("serializer", String.class);
  }

  /**
  * Gets the maxRetries value from the configuration.
  * Attribute Name: maxRetries
  * Description: The number of retries to make for sending a message to a remote websocket endpoint. A value greater than 0 is advised. Otherwise, a web socket timeout can result in a dropped message
  * Default Value: 1
  * @return the maxRetries
  */
  public Integer getMaxRetries() {
    return config.getOptionalValue("maxRetries", Integer.class)
     .orElse(Integer.valueOf("1"));
  }

  /**
  * Gets the jitter value from the configuration.
  * Attribute Name: jitter
  * Description: Configures the random factor when using back-off with maxAttempts > 1
  * Default Value: 0.5
  * @return the jitter
  */
  public Double getJitter() {
    return config.getOptionalValue("jitter", Double.class)
     .orElse(Double.valueOf("0.5"));
  }

  /**
  * Gets the delay value from the configuration.
  * Attribute Name: delay
  * Description: Configures a back-off delay between attempts to send a request. A random factor (jitter) is applied to increase the delay when several failures happen.
  * @return the delay
  */
  public Optional<String> getDelay() {
    return config.getOptionalValue("delay", String.class);
  }

  public void validate() {
    super.validate();
    getUrl();
  }
}
